//+------------------------------------------------------------------+
//|                                                        timer.mq4 |
//|                                      Copyright  2016, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//|            MQL's OOP Notes: Object hierarchies and serialization |
//|                        https://www.mql5.com/en/blogs/post/682501 |
//+------------------------------------------------------------------+

#property strict

class OutputStream;
class InputStream;

class Object
{
  public:
    virtual string toString() const {return classname() + "#" + StringFormat("%d", &this);}
  
    virtual bool save(OutputStream &out) const {return false;}
    virtual bool load(InputStream &in) {return false;}

    virtual long version() const {return 0;}
    virtual string classname() const = 0;    
    virtual long signature() const {return 0;}
    
    virtual int compare(const Object &other, const double customValue = 0) const
    {
      return (int)(signature() - other.signature());
    }
};


class OutputStream
{
  public:
    // simple base stubs
    template<typename T>
    bool write(const T t)
    {
      return writeString((string)t);
    }
    template<typename T>
    bool write(const T *t)
    {
      return t.save(this);
    }
    
    // customized part
    virtual bool writeBoolean(const bool b) {return false;}
    virtual bool writeByte(const uchar c){ return false;}
    virtual bool writeInt(const int i) {return false;}
    virtual bool writeLong(const long l) {return false;}
    virtual bool writeDouble(const double d) {return false;}
    virtual bool writeDatetime(const double t) {return false;}
    
    // must be defined in descendants
    virtual bool writeString(const string s) = 0;
};

class LogOutputStream: public OutputStream
{
  public:
    virtual bool writeString(const string s)
    {
      Print(s);
      return true;
    }
};

class TextFileWriter: public OutputStream
{
  private:
    string name; 
    int handle;
     
  public:
    TextFileWriter(const string filename)
    {
      name = filename;
      handle = INVALID_HANDLE;
      open();
    }

    ~TextFileWriter()
    {
      close();
    }
    
    void close()
    {
      if(handle != INVALID_HANDLE)
      {
        FileClose(handle);
        handle = INVALID_HANDLE;
      }
    }
    
    void open()
    {
      if(handle == INVALID_HANDLE)
      {
        handle = FileOpen(name, FILE_WRITE | FILE_CSV | FILE_SHARE_READ | FILE_SHARE_WRITE, ',');
        if(handle == INVALID_HANDLE)
        {
          Print("TextFileWriter: can't open ", name);
        }
      }
    }

    virtual bool writeString(const string s)
    {
      return FileWrite(handle, s) != 0;
    }
}; 

class BinaryFileWriter: public OutputStream
{
  // TODO
};

class InputStream
{
  public:
    virtual bool readBoolean() = 0;
    virtual uchar readByte() = 0;
    virtual int readInt() = 0;
    virtual long readLong() = 0;
    virtual string readString() = 0;
    virtual double readDouble() = 0;
    virtual datetime readDatetime() = 0;
};

class TextFileReader : public InputStream
{
  private:
     int handle;
     
  public:
    TextFileReader(const string filename)
    {
      handle = FileOpen(filename, FILE_READ | FILE_CSV | FILE_SHARE_READ | FILE_SHARE_WRITE, ',');
      if(handle == INVALID_HANDLE)
      {
        Print("TextFileReader: can't open ", filename);
      }
    }

    ~TextFileReader()
    {
      if(handle != INVALID_HANDLE)
      {
        FileClose(handle);
      }
    }
    
    virtual bool readBoolean()
    {
      return FileReadBool(handle);
    }
    
    virtual uchar readByte()
    {
      return (uchar)FileReadNumber(handle);
    }
    
    virtual int readInt()
    {
      return (int)FileReadNumber(handle);
    }
    
    virtual long readLong()
    {
      return (long)FileReadNumber(handle);
    }
    
    virtual string readString()
    {
      return FileReadString(handle);
    }
    
    virtual double readDouble()
    {
      return FileReadNumber(handle);
    }

    virtual datetime readDatetime()
    {
      return FileReadDatetime(handle);
    }

};

class Pair : public Object
{
  private:
    static const long v;
    
  public:
    double x, y;
    Pair(const double _x, const double _y): x(_x), y(_y) {}
    Pair(const Pair &p2)
    {
      x = p2.x;
      y = p2.y;
    }
    Pair(const Pair *p2)
    {
      if(CheckPointer(p2) != POINTER_INVALID)
      {
        x = p2.x;
        y = p2.y;
      }
      else
      {
        x = y = 0.0;
      }
    }
    Pair(): x(0), y(0) {}
    
    virtual string classname() const
    {
      return typename(this);
    }
    virtual long version() const
    {
      return v;
    }
    
    virtual bool save(OutputStream &out) const
    {
      out.write(classname() + (string)version());
      out.write(x);
      out.write(y);
      return true;
    }
    
    virtual bool load(InputStream &in)
    {
      string type = in.readString();
      if(type != classname() + (string)version())
      {
        Print("Class mismatch:", type, "<>", classname() + (string)version());
        return false;
      }
      x = in.readDouble();
      y = in.readDouble();
      
      return true;
    }
    
    virtual int compare(const Pair &other, const double customValue = 0) const
    {
      return !(x == other.x && y == other.y);
    }
};

const long Pair::v = 1;

class Example : public Object
{
  private:
    int index;
    double data;
    string text;
    datetime dt;
    Pair pair;
    
  public:
    Example(const int i, const double d, const string t, const datetime m, Pair &p): index(i), data(d), text(t), dt(m), pair(p) {}
    Example(const int i, const double d, const string t, const datetime m): index(i), data(d), text(t), dt(m) {}
    
    Example(): index(0), data(0), text(NULL), dt(0) {}
    
    virtual string classname() const
    {
      return typename(this);
    }
    
    virtual bool save(OutputStream &out) const
    {
      out.write(classname() + (string)version());
      out.write(index);
      out.write(data);
      out.write(text);
      out.write(dt);
      out.write(&pair); // pair.save(out);
      return true;
    }
    
    virtual bool load(InputStream &in)
    {
      string type = in.readString();
      if(type != classname() + (string)version())
      {
        Print("Class mismatch:", type, "<>", classname() + (string)version());
        return false;
      }
      index = in.readInt();
      data = in.readDouble();
      text = in.readString();
      dt = in.readDatetime();
      pair.load(in);
      
      return true;
    }
    
    virtual int compare(const Example &other, const double customValue = 0) const
    {
      return !(index == other.index && data == other.data && text == other.text && (pair.compare(other.pair, customValue) == 0));
    }
    
    void applyPair(const float x, const float y)
    {
      pair.x = x;
      pair.y = y;
    }
    
    bool operator==(const Example &other) const
    {
      return !compare(other);
    }
  
};

void OnStart()
{
  LogOutputStream log;  // this is just for debug purpose

  Example e1(1, 10.5, "text", TimeLocal());
  Print(e1.toString());
  e1.save(log);
  
  TextFileWriter tfw("file1.txt");
  e1.save(tfw);         // serialize object to the text file
  tfw.close();          // release file handle
  
  TextFileReader tfr("file1.txt");
  Example e2;
  e2.applyPair(-1, -2); // apply some changes, just to make sure load will override them
  e2.load(tfr);         // deserialize the text into another object instance
  Print(e2.toString());
  e2.save(log);
  
  Print("Is equal after restore:", (e1 == e2));      // true, obects are identical
  e2.applyPair(5, 6);
  Print("Is equal after modification:", (e1 == e2)); // false
}